/*
 * Access a filesystem.
 *
 * Copyright (C) 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/types.h>
#include <l4/util/util.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <fsclient/client.h>
#include <fsclient/fsdesc.h>
#include <fsserver/accessor.h>
#include <fsserver/file_resource.h>
#include <fsserver/file_resource_server.h>
#include <fsserver/user_filesystem_resource.h>
#include <systypes/fcntl.h>
#include "fslib.h"



/* Filesystem accessor. */

class Fs_accessor : public FileAccessor
{
  /* Interaction with the block device region. */

  file_descriptor_t *_region;

  /* File location. */

  size_t _location;

public:
  Fs_accessor(file_descriptor_t *region, size_t location, size_t size)
  : FileAccessor(size), _region(region), _location(location)
  {
  }

  long flush()
  {
    return L4_EOK;
  }

  size_t fstat(char *, size_t)
  {
    return 0;
  }

  size_t read(size_t start, char *buffer, size_t length)
  {
    /* Get the desired position of the data for reading. */
    /* NOTE: Assuming contiguous data. Should check for file extent. */

    size_t position = _location + start;

    /* Ensure that data is available in the region. */

    long err = client_ensure_read_position(_region, position);
    if (err)
      return 0;

    return client_read(_region, (void *) buffer, length);
  }

  size_t write(size_t, char *, size_t)
  {
    return 0;
  }
};



/* User filesystem server. */

class Fs_userfilesystem : public UserFilesystemResource
{
private:
  file_descriptor_t *_region;
  const char *_devname;

public:
  explicit Fs_userfilesystem(file_descriptor_t *region, const char *devname)
  : UserFilesystemResource(), _region(region), _devname(devname)
  {
  }

  long open_object_server(fs_object_t *fsobj, int flags,
                          Accessor **accessor,
                          ResourceServer **server)
  {
    (void) flags;

    /* Obtain the object's path. */

    const char *path = (const char *) fsdesc_get_name(fsobj);

    printf("path = %s\n", path);

    if (path == NULL)
      return -L4_EINVAL;

    /* Find the object. If it cannot be located, return an error. */

    client_seek(_region, 0, SEEK_SET);

    entry_t entry;

    if (!find_entry_recursively(_region, path, &entry))
      return -L4_EIO;

    /* Create an accessor for the object. */

    Fs_accessor *fs_accessor = new Fs_accessor(_region, entry.location, entry.metadata.size);

    /* Return the accessor and object server. */

    *accessor = fs_accessor;
    *server = new FileServer(new FileResource(fs_accessor, *fsobj));

    return L4_EOK;
  }
};



int main(int argc, char *argv[])
{
  file_descriptor_t *region;
  char *devname;

  if (argc < 2)
  {
    printf("Need a block device capability or a file accessible via the fs capability.\n");
    return 1;
  }

  /* Obtain a region of a block device, potentially via the filesystem. */

  devname = argv[1];

  if (l4_is_invalid_cap(l4re_env_get_cap("fs")))
    region = client_open_device(devname, O_RDWR, 0);
  else
    region = client_open(devname, O_RDWR, 0);

  if (region == NULL)
  {
    printf("Could not obtain region for %s.\n", devname);
    return 1;
  }

  /* Initialise and register a new server object. */

  Fs_userfilesystem filesystem(region, devname);
  UserFilesystemServer server(&filesystem);

  if (server.bind("export"))
  {
    printf("Could not bind thread.\n");
    return 1;
  }

  /* Enter the IPC server loop. */

  printf("Starting simple filesystem...\n");
  server.start();
  return 0;
}
